# -*- coding: UTF-8 -*-

# TODO: Incorporate LMDB-based dataloading

"""Utilities for constructing the appropriate set of torch Datasets & DataLoaders.

Functions "prepare_dataset(...)" and "prepare_dataloader(...)" construct a set
of torch Datasets and DataLoaders respectively based off of configurations
specified in data_config.py and config.py; both of these modules have to be
run atleast once (see their respective docstrings for instructions on running
them) before "prepare_dataset(...)" and "prepare_dataloader(...)" can be run.

  Typical usage examples:

  train_ds, valid_ds = prepare_dataset( data_config )
  train_dl, valid_dl, z_valid_dl = prepare_dataloader( config, data_config, train_ds, valid_ds )

The output DataLoaders can then used as inputs into the train() method of the
appropriate Learner (which was determined when you ran config.py), as is done,
for example, when running train.py on the command-line.

The functions in this module can also be used externally as data-construction
utilities.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from .latent_utils import gen_rand_latent_vars, concat_rand_classes_to_z
from _int import get_current_configuration, \
                 TRAINING_SET_DIRNAME, VALID_SET_DIRNAME, TORCHVISION_CATERED_DATASETS
from data_config import get_dataset_img_extension

from functools import partial
from pathlib import Path

import numpy as np
import torch
from torchvision.datasets.folder import *
from torchvision import datasets
from torch.utils.data import TensorDataset, DataLoader, RandomSampler, SequentialSampler, BatchSampler

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

# torch Dataset Utilities:
# ------------------------

def prepare_dataset( data_config ):
  """Construct torch Dataset(s) based on configuration specified when data_config.py was last ran."""
  dataset_maker = partial(
    make_torchvision_dataset,
    dataset = data_config.dataset,
    dataset_dir = data_config.dataset_dir,
    ds_transforms = data_config.ds_transforms
  )

  valid_ds = None
  if data_config.is_torchvision_catered_ds or data_config.ds_ext in IMG_EXTENSIONS:
    train_ds = dataset_maker( is_training_set = True )
    if data_config.include_valid_set:
      valid_ds = dataset_maker( is_training_set = False )
  else:  # use DatasetFolderSingleClass
    train_ds = dataset_maker( is_training_set = True,
                              loader = data_config.ds_loader, extensions = ( data_config.ds_ext, ) )
    if data_config.include_valid_set:
      valid_ds = dataset_maker( is_training_set = False,
                                loader = data_config.ds_loader, extensions = ( data_config.ds_ext, ) )

  return train_ds, valid_ds

# ............................................................................ #

# TODO: Implement class-conditioning option (Mirza & Osindero, 2014) for discriminator input.
def make_torchvision_dataset( dataset, dataset_dir, is_training_set = True, ds_transforms = None, *args, **kwargs ):
  if not isinstance( dataset_dir, ( str, Path, ) ):
    raise TypeError( '"dataset_dir" must be of type "str" or "pathlib.Path"' )
  dataset_dir = str( dataset_dir )

  dataset_title = dataset.casefold()
  # TODO: Add more torchvision-catered datasets:
  if dataset_title in TORCHVISION_CATERED_DATASETS['LSUN Bedrooms'][0]:
    config = get_current_configuration( 'config', raise_exception = False )
    if config is not None:
      if config.res_dataset > 256:
        message = f'WARNING: config.res_dataset currently set to {config.res_dataset},' + \
                  f' but recommended to set --res_dataset to 256 or below when using LSUN Bedrooms!'
        raise RuntimeWarning( message )
    classes_categories = TORCHVISION_CATERED_DATASETS['LSUN Bedrooms'][1]
    classes = classes_categories[0] if is_training_set else classes_categories[1]
    dataset = datasets.LSUN(
      root = dataset_dir, classes = classes, transform = ds_transforms
    )
  elif dataset_title in TORCHVISION_CATERED_DATASETS['CIFAR-10'][0]:
    config = get_current_configuration( 'config', raise_exception = False )
    if config is not None:
      if config.res_dataset > 32:
        message = f'WARNING: config.res_dataset currently set to {config.res_dataset},' + \
                  f' but recommended to set --res_dataset to 32 or below when using CIFAR-10!'
        raise RuntimeWarning( message )
    classes_categories = TORCHVISION_CATERED_DATASETS['CIFAR-10'][1]
    classes = classes_categories[0] if is_training_set else classes_categories[1]
    dataset = datasets.CIFAR10(
      root = dataset_dir, train = classes, transform = ds_transforms, download = True
    )
  else:
    # Custom Datasets:
    imgs_dirname = TRAINING_SET_DIRNAME if is_training_set else VALID_SET_DIRNAME
    if get_dataset_img_extension( dataset_dir + '/' + imgs_dirname ) in IMG_EXTENSIONS:
      dataset = ImageFolderSingleClass( root = dataset_dir, category = imgs_dirname, transform = ds_transforms )
    else:
      if 'loader' not in kwargs or kwargs[ 'loader' ] is None:
        raise ValueError( 'Generic DatasetFolderSingleClass requires input `loader:callable` auxiliary argument.' )
      raise RuntimeWarning( '\nUsing generic DatasetFolderSingleClass...\n' + \
                            ' WARNING: The input "loader" argument for this class that acts on your data must output an object that' + \
                                       ' can be converted into a PIL Image object (e.g. an object of type `np.uint8` or `torch.uint8`).\n' )
      loader = kwargs.pop( 'loader' )
      dataset = DatasetFolderSingleClass( root = dataset_dir, category = imgs_dirname,
                                          loader = loader, transform = ds_transforms, **kwargs )
  return dataset

class DatasetFolderSingleClass( datasets.DatasetFolder ):
  """torchvision's DatasetFolderClass but with the ability to just use 1 class."""
  def __init__( self, root, category, loader, extensions = None, transform = None,
                target_transform = None, is_valid_file = None ):
    super( datasets.DatasetFolder, self ).__init__( root, transform = transform,
                                                    target_transform = target_transform )
    # classes, class_to_idx = self._find_classes(self.root)
    classes = [ category ]; class_to_idx = { category: 0 }
    samples = make_dataset(self.root, class_to_idx, extensions, is_valid_file)
    if len(samples) == 0:
      raise ( RuntimeError( 'Found 0 files in subfolders of: ' + self.root + '\n'
                            'Supported extensions are: ' + ','.join( extensions ) )
            )

    self.loader = loader
    self.extensions = extensions

    self.classes = classes
    self.class_to_idx = class_to_idx
    self.samples = samples
    self.targets = [s[1] for s in samples]

class ImageFolderSingleClass( DatasetFolderSingleClass ):
  """torchvision's ImageFolderClass but with the ability to just use 1 class."""
  def __init__( self, root, category, transform = None, target_transform=None,
                loader = default_loader, is_valid_file = None ):
    super( ImageFolderSingleClass, self ).__init__( root, category, loader,
                                                    IMG_EXTENSIONS if is_valid_file is None else None,
                                                    transform = transform,
                                                    target_transform = target_transform,
                                                    is_valid_file = is_valid_file
                                                  )
    self.imgs = self.samples

# ............................................................................ #

# TODO: Implement more types of loaders
def npy_loader( path ):
  return torch.from_numpy( np.load( path ) ).squeeze()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

# torch DataLoader Utilities:
# ---------------------------

def prepare_dataloader( config, data_config, train_ds, valid_ds = None ):
  """Construct torch DataLoader(s) based on configuration specified when config.py and data_config.py was last ran."""
  if data_config.include_valid_set:
    # DataLoader for latent vectors for validation:
    z_valid = gen_rand_latent_vars(
                num_samples = len( valid_ds ),
                length = config.len_latent,
                distribution = config.latent_distribution,
                device = 'cpu'
              )
    if config.class_condition or config.use_auxiliary_classifier:
        z_labels = torch.randint( 0, config.num_classes, ( len( z_valid ), 1, ),
                                  dtype = torch.int64, device = 'cpu' )
    if config.class_condition:
        z_valid = concat_rand_classes_to_z( z = z_valid, num_classes = config.num_classes,
                                            z_labels = z_labels, device = 'cpu' )
    if config.class_condition or config.use_auxiliary_classifier:
        z_labels.squeeze_()
        z_valid_ds = TensorDataset( z_valid, z_labels )
    else:
        z_valid_ds = TensorDataset( z_valid )

  # DataLoader(s) for training data:
  dataloader = configure_dataloader_for_hardware( num_workers = config.num_workers, pin_memory = config.pin_memory )

  valid_dl = None; z_valid_dl = None
  if config.model == 'ResNet GAN':
    train_dl = dataloader( dataset = train_ds, batch_size = config.batch_size,
                           shuffle = True, drop_last = True )
    if data_config.include_valid_set:
      valid_dl = dataloader( dataset = valid_ds, batch_size = config.batch_size,
                             shuffle = False, drop_last = False )
      z_valid_dl = dataloader( dataset = z_valid_ds, batch_size = config.batch_size,
                               shuffle = False, drop_last = False )
  elif config.model in ( 'ProGAN', 'StyleGAN', ):
    train_batch_sampler = BatchSampler( sampler = RandomSampler( data_source = train_ds ),
                                        batch_size = config.bs_dict[ config.init_res ], drop_last = True )
    train_dl = dataloader( dataset = train_ds, batch_sampler = train_batch_sampler )
    if data_config.include_valid_set:
      valid_batch_sampler = BatchSampler( sampler = SequentialSampler( data_source = valid_ds ),
                                          batch_size = config.bs_dict[ config.init_res ], drop_last = False )
      valid_dl = dataloader( dataset = valid_ds, batch_sampler = valid_batch_sampler )
      z_valid_batch_sampler = BatchSampler( sampler = SequentialSampler( data_source = z_valid_ds ),
                                            batch_size = config.bs_dict[ config.init_res ], drop_last = False )
      z_valid_dl = dataloader( dataset = z_valid_ds, batch_sampler = z_valid_batch_sampler )
  else:
    message = f'Model type {config.model} not supported, thus could not construct torch DataLoader.'
    raise ValueError( message )

  return train_dl, valid_dl, z_valid_dl

# ............................................................................ #

def configure_dataloader_for_hardware( num_workers, pin_memory ):
  dataloader = partial(
    DataLoader,
    num_workers = num_workers,
    pin_memory = pin_memory
  )
  return dataloader

class DataLoaderEx( DataLoader ):
  """Customize your own DataLoader."""
  def __init__( self, dataset , bs ):
    super( DataLoaderEx, self ).__init__( dataset, bs )
    raise NotImplementedError( 'Class "DataLoaderEx" not yet implemented.' )


# -*- coding: UTF-8 -*-

"""Used internally.
"""

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

import os
import argparse
import copy
from pathlib import Path
import pickle

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

# Currently Supported Learners:
SUPPORTED_LEARNERS = ( 'GANLearner', 'ProGANLearner', 'StyleGANLearner', )

# Currently Supported Architectures:
SUPPORTED_ARCHS = {
  'Generators'        : ( 'Generator32PixResnet', 'Generator64PixResnet',
                          'ProGenerator', 'StyleGenerator',
  ),
  'Discriminators'    : ( 'Discriminator32PixResnet', 'Discriminator64PixResnet',
                          'ProDiscriminator', 'StyleDiscriminator',
  ),
  'AC Discriminators' : ( 'DiscriminatorAC32PixResnet', 'DiscriminatorAC64PixResnet', )
}

# For custom datasts only:
TRAINING_SET_DIRNAME = 'train'
VALID_SET_DIRNAME = 'valid'

# For torchvision-catered datasets only:
TORCHVISION_CATERED_DATASETS = {
  'LSUN Bedrooms' : (
                       ( 'lsun bedrooms', 'lsun-bedrooms', 'lsunbedrooms', ),
                       ( [ 'bedroom_train' ], [ 'bedroom_val' ], ),
  ),
  'CIFAR-10'      : (
                       ( 'cifar-10', 'cifar10', 'cifar 10', ),
                       ( True, False, ),
  )
}

FMAP_SAMPLES = 3
RES_INIT = 4

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

class NotConfiguredError( Exception ):
  """Raised when either config.py or data_config.py has not been run."""
  pass

def get_current_configuration( cfg, raise_exception = True ):

  try:
    with open( str( Path.home()/'.configs_dir.txt' ), 'rb' ) as f:
      configs_dir = f.readline().decode( 'utf8' ).strip()
  except FileNotFoundError:
    if raise_exception:
      message = f'Please run config.py or data_config.py atleast once before running this function.'
      raise NotConfiguredError( message )
    else:
      return None

  if cfg is 'config':
    _pickled_file = configs_dir + '/.config.p'
    module = 'config.py'
  elif cfg is 'data_config':
    _pickled_file = configs_dir + '/.data_config.p'
    module = 'data_config.py'
  else:
    raise ValueError( "Input configuration does not exist. Options are 'config' and 'data_config'." )

  try:
    with open( _pickled_file, 'rb' ) as f:
      return pickle.load( f )
  except FileNotFoundError:
    if raise_exception:
      message = f'Please run {module} atleast once in order to obtain your desired configuration.'
      raise NotConfiguredError( message )
    else:
      return None

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

# https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse
def str2bool( v ):
  """Enables intuitive boolean keyword argument specifications with argparse."""
  if isinstance( v, bool ):
    return v
  if v.casefold() in ( 'yes', 'true', 't', 'y', '1' ):
    return True
  elif v.casefold() in ( 'no', 'false', 'f', 'n', '0' ):
    return False
  else:
    raise argparse.ArgumentTypeError( 'Boolean value expected.' )

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

class LearnerConfigCopy( object ):
  def __init__( self, config, learner_class:str,
                nonredefinable_attrs:tuple, redefinable_from_learner_attrs:tuple ):
    assert ( isinstance( config, argparse.Namespace ) and 'model' in config.__dict__ )
    super( LearnerConfigCopy, self ).__init__()

    object.__setattr__( self, '__dict__', copy.deepcopy( config.__dict__ ) )

    if learner_class is 'GANLearner':
      self.__dict__[ 'model_name' ] = 'resnetgan'
    elif learner_class is 'ProGANLearner':
      self.__dict__[ 'model_name' ] = 'progan'
    elif learner_class is 'StyleGANLearner':
      self.__dict__[ 'model_name' ] = 'stylegan'
    else:
      message = f"Input learner_class argument set equal to {learner_class}." + \
                f" But currently, learner_class can only be\n" + \
                f"one of: [ '" + "', '".join( SUPPORTED_LEARNERS ) + "' ]"
      raise ValueError( message )

    self.__dict__[ 'learner_class' ] = learner_class
    self.__dict__[ '_nonredefinable_attrs' ] = nonredefinable_attrs
    self.__dict__[ '_redefinable_from_learner_attrs' ] = redefinable_from_learner_attrs

  def __setattr__( self, name, value ):
    if name in self._nonredefinable_attrs:
      message0 = f"{self.learner_class}().config.{name} attribute cannot be changed once {self.learner_class} is instantiated.\n"
      if name == 'model':
        message1 = f"Instead, please run 'python config.py {value}' on the command-line and then instantiate a new {self.learner_class}."
      else:
        message1 = f"Instead, please run 'python config.py {self.model_name} --{name}={value}' on the command-line and then instantiate a new {self.learner_class}."
      message = message0 + message1
      raise AttributeError( message )
    elif name in self._redefinable_from_learner_attrs:
      message = f"{self.learner_class}().config.{name} attribute cannot be changed.\n" + \
                f" Instead, please change {self.learner_class}().{name} to implement this change in the {self.learner_class} instance,\n" + \
                f" while {self.learner_class}().config.{name} will remain equal to its value when the {self.learner_class} was first initialized."
      raise AttributeError( message )
    else:
      super( LearnerConfigCopy, self ).__setattr__( name, value )

  def __str__( self ):
    print_obj = ''
    for k, v in vars( self ).items():
      if k not in ( '_nonredefinable_attrs', '_redefinable_from_learner_attrs', 'model_name', 'learner_class', ):
        print_obj += f'  {k}: {v}\n'
    return print_obj[:-1]


# -*- coding: UTF-8 -*-

"""Various utilities for all aspects of GAN training and evaluation.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

import os; import sys
sys.path.append(
  os.path.abspath( os.path.join( os.path.dirname( __file__ ), '..' ) )
)

from . import latent_utils
from . import data_utils
from . import initializer
from . import custom_layers
from . import backprop_utils

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #


# -*- coding: UTF-8 -*-

"""Configures package for the user-specified model type and hyperparameters.

Run 'python config.py [resnet, progan, stylegan]' on the command-line to
configure (or re-configure) this package for the model type you just specified,
and optionally specify any set of '--[keyword argument]' to override any default
hyperparameters (the defaults work well).

  Typical usage examples:

  $ python config.py stylegan  [OR]
  $ python config.py progan --dev=cuda --enable_cudnn_autotuner --init_res=8

After running this module, you will be able to instantiate (or re-instantiate)
a Learner corresponding to the model type & hyperparameters that you just
configured (or re-configured) for, at any later time.

If you would like to see a list of what each argument does,
run 'python config.py [resnet, progan, stylegan] -h' on the command-line.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from _int import str2bool, get_current_configuration

import os
import sys
import argparse
from functools import partial
# import time
from pathlib import Path
import pickle

import numpy as np
import torch
from torchvision import transforms

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

DEV = 'cuda' if torch.cuda.is_available() else 'cpu'

BS = 64
NIMG_TRANSITION = 600000

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

if __name__ == '__main__':

  parser = argparse.ArgumentParser( prog = 'GAN Lab', description = 'StyleGAN, ProGAN, and ResNet GANs (model configuration)' )

  # ---------------------------------------------------------------------------- #

  # POSITIONAL (REQUIRED) ARGUMENTS:
  # --------------------------------

  parser.add_argument( 'model', type = str, help = 'which of the 3 models (ResNet GAN, ProGAN, or StyleGAN) to use' )


  if len( sys.argv ) < 2:
    raise TypeError( 'config.py requires atleast 1 argument (the model type), and then optional keyword arguments (prefixed with "--").\n' + \
                     ' Example: "$ python config.py stylegan --init_res=8"' )

  if sys.argv[1].casefold() in ( 'resnet', 'res-net', 'res net', 'resnet gan',
                                 'res-net gan', 'res net gan', 'resnetgan', ):
    _model_type = 'ResNet GAN'
  elif sys.argv[1].casefold() in ( 'progan', 'pro-gan', 'pro gan', 'pggan', 'pg-gan', 'pg gan',
                                   'progressivegan', 'progressive-gan', 'progressive gan', ):
    _model_type = 'ProGAN'
  elif sys.argv[1].casefold() in ( 'stylegan', 'style-gan', 'style gan', ):
    _model_type = 'StyleGAN'
  else:
    raise ValueError( 'Invalid model input. The GAN Lab currently only' + \
                      ' supports ResNet GAN, Progressive GAN, or StyleGAN.' )

  # ---------------------------------------------------------------------------- #

  # KEYWORD (OPTIONAL) ARGUMENTS:
  # -----------------------------

  parser.add_argument(
    '--dev',
    type = str.casefold,
    default = DEV,
    choices = [ 'cpu', 'cuda' ],
    help = 'device for memory allocation for the generator and discriminator'
  )
  # TODO: Implement multi-GPU support
  parser.add_argument( '--n_gpu', default = 1, type = int, help = "number of GPUs to use; ignored if config.dev is 'cpu'" )
  # TODO: Implement dynamic batch size control of _MultiProcessDataLoaderIter instances (instantiated when num_workers > 0).
  parser.add_argument( '--enable_cudnn_autotuner', type = str2bool, nargs = '?', const = True, default = False, \
    help = "whether to enable CuDNN auto-tuner to find best algorithm to use for your hardware; usually leads to faster runtime" + \
           " when input size for neural network doesn't vary, as then cuddn will not need to benchmark every time input size changes" )

  parser.add_argument( '--random_seed', type = int, default = -1, \
    help = 'seed for RNG operations such as latent vector sampling, noise, etc.; making this a non-negative integer (instead of -1)' + \
           ' leads to deterministic output, which may be desired if a goal is reproducibility; making this -1 randomly samples the seed instead' )

  parser.add_argument( '--gen_bs_mult', type = int, default = 1, \
    help = 'multiplicative factor for training generator with more data per batch than discriminator;' + \
           ' some datasets (such as CIFAR-10 in ResNet GAN) benefit from this' )

  parser.add_argument( '--num_gen_iters', type = int, default = 1, \
    help = 'number of generator iterations in one main iteration through generator and discriminator' )

  parser.add_argument( '--loss', type = str.casefold, default = 'wgan', choices = [ 'wgan', 'nonsaturating', 'minimax' ], \
    help = 'type of loss function to use for the generator and discriminator' )
  parser.add_argument( '--gradient_penalty', type = str.casefold, default = 'wgan-gp', choices = [ None, 'wgan-gp', 'r1', 'r2' ], \
    help = "type of gradient regularizer to use; options 'R1' and 'R2' are the zero-centered gradient regularizers from" + \
           " `Which Training Methods for GANs do actually Converge?` (Mescheder et al., 2018)" )
  parser.add_argument( '--lda', type = float, default = 10., \
    help = 'gradient penalty multiplicative coefficient (for the `wgan-gp`, `r1`, and `r2` --gradient_penalty options above)' )
  parser.add_argument( '--gamma', type = float, default = 1., \
    help = "gradient penalty degree of Lipschitz constraint; multiclass datasets may benefit from gamma > 1. to help mitigate ghosting;" + \
           " ignored if --gradient_penalty is not set to 'wgan-gp'" )

  parser.add_argument( '--lr_sched_custom', type = str.casefold, default = None, \
    help = 'learning rate scheduler for custom --lr_sched argument; otherwise ignored; this should be a lambda function string' )

  parser.add_argument( '--optimizer', type = str.casefold, default = 'adam', \
    choices = [ 'adam', 'rmsprop', 'momentum', 'sgd' ], help = 'which optimizer to train with' )
  parser.add_argument( '--beta1', type = float, default = 0., help = 'momentum EWMA decay coefficient (default value recommended for Adam)' )
  parser.add_argument( '--eps', type = float, default = 1.e-8, help = 'to prevent division by 0 (default value recommended for Adam)' )
  parser.add_argument( '--wd', type = float, default = 0., help = 'weight decay multiplicative coefficient' )

  parser.add_argument( '--align_corners', type = str2bool, nargs = '?', const = True, default = False, \
    help = 'only applies for bilinear; if `True`, aligns corner pixels of input and output tensors of interpolation' )
  parser.add_argument( '--model_upsample_type', type = str.casefold, default = 'nearest', choices = [ 'nearest', 'bilinear' ], \
    help = 'type of upsampling that the generator conducts to go from latent space to image space' )
  parser.add_argument( '--model_downsample_type', type = str.casefold, default = 'average', choices = [ 'nearest', 'average', 'box', 'bilinear' ], \
    help = 'type of downsampling that the discriminator conducts to go from image space to final feature space' )

  parser.add_argument( '--latent_distribution', type = str.casefold, default = 'normal', choices = [ 'normal', 'uniform' ], \
    help = 'to sample elements of latent vector from normal distribution or uniform distribution' )

  # TODO: Implement class-conditioning of discriminator when not using auxiliary classifier. Also implement more modern forms of the latter.
  parser.add_argument( '--num_classes', type = int, default = 0, \
    help = 'number of classes used for class-conditioning (if `True` below) and/or auxiliary classifier (if `True` below);' + \
           ' if both are `False` below, the model is unsupervised' )
  parser.add_argument( '--class_condition', type = str2bool, nargs = '?', const = True, default = False, \
    help = 'whether to use class conditioning in generator as well as discriminator (the latter only if not using auxiliary classifier (below))' )
  parser.add_argument( '--use_auxiliary_classifier', type = str2bool, nargs = '?', const = True, default = False, \
    help = 'whether to use auxiliary classifier at the end of discriminator (ACGAN)' )
  parser.add_argument( '--ac_disc_scale', type = float, default = 1., \
    help = 'multiplicative factor applied to ACGAN (if --use_auxiliary_classifier set to `True` above) loss when training the discriminator' )
  parser.add_argument( '--ac_gen_scale', type = float, default = .1, \
    help = 'multiplicative factor applied to ACGAN (if --use_auxiliary_classifier set to `True` above) loss when training the generator' )

  # NOTE: All of the below 6 arguments will be ignored if --include_valid_set in data_config.py is set to `False`:
  parser.add_argument( '--num_iters_valid', type = int, default = 1000, \
    help = 'number of main iterations until evaluating metrics again using validation set during training;' + \
           ' ignored if user does not wish to use a validation set during training (see data_config.py for details)' )
  parser.add_argument( '--metrics_dev', type = str.casefold, default = 'cpu', choices = [ 'cpu', 'cuda' ], \
    help = 'device for memory allocation when computing metrics using the validation set' )
  parser.add_argument( '--gen_metrics', type = list, default = [ 'generator loss', 'fake realness', 'image grid' ], \
    choices = [ 'generator loss', 'fake realness', 'image grid' ], help = 'what metrics to evaluate for the generator' )
  parser.add_argument( '--disc_metrics', type = list, default = [ 'discriminator loss', 'fake realness', 'real realness' ], \
    choices = [ 'discriminator loss', 'fake realness', 'real realness' ], help = 'what metrics to evaluate for the discriminator' )
  parser.add_argument( '--img_grid_sz', type = int, default = 4, \
    help = 'size of image grid saved every time metric evaluation is performed on the validation set;' + \
           ' only utilized if image grid is specified in gen_metrics (2 above)' )
  parser.add_argument( '--img_grid_show_labels', type = str2bool, nargs = '?', const = True, default = True, \
    help = 'whether to show the ground truth labels on the image grid for supervised models; ignored if unsupervised model' + \
           ' (i.e. if both --class_condition and --use_auxiliary_classifier above are `False`)' )

  parser.add_argument( '--save_samples_dir', type = Path, default = os.path.abspath( os.path.dirname( __file__ ) ) + '/samples/', \
    help = "root directory where all saved samples are stored (e.g. the image grid from the 'image grid' metric (see --gen_metrics above))" )

  parser.add_argument( '--num_iters_save_model', type = int, default = 1000, help = "number of main iterations until model is saved again during training" )
  parser.add_argument( '--save_model_dir', type = Path, default = os.path.abspath( os.path.dirname( __file__ ) ) + '/models/', \
    help = "root directory where the model is saved by default, such as every config.num_iters_save_model iterations (see --num_iters_save_model above)" )

  parser.add_argument( '--num_workers', type = int, default = 0, \
    help = 'number of subprocesses to use for data loading; 0 will make data load in main process' )
  parser.add_argument( '--pin_memory', type = str2bool, nargs = '?', const = True, default = True if DEV == 'cuda' else False, \
    help = 'whether the dataloader will copy torch.Tensors into CUDA pinned memory, thus enabling faster data transfer; set to `False` if not using CUDA' )

  # ............................................................................ #

  # ResNet GAN Keyword Arguments:
  if _model_type is 'ResNet GAN':
    parser.add_argument( '--batch_size', type = int, default = BS, help = 'base batch size of training/generated data' )

    parser.add_argument( '--num_main_iters', type = int, default = 300000, help = 'number of main iterations of training generator and discriminator' )
    parser.add_argument( '--num_disc_iters', type = int, default = 5, \
      help = "number of discriminator iterations in one main iteration through generator and discriminator;" + \
             " suggested to train to optimality so that minimizing generator's loss is the same as minimizing the JS-divergence for many types of loss functions" )

    parser.add_argument( '--lr_base', type = float, default = .0001, help = 'base (max) learning rate; can vary depending on dataset and model' )
    parser.add_argument( '--lr_sched', type = str.casefold, default = None, \
      choices = [ None, 'linear decay', 'custom' ], help = 'learning rate scheduler type' )

    parser.add_argument( '--beta2', type = float, default = .9, help = 'variance/RMSprop EWMA decay coefficient (default recommended for Adam)' )

    parser.add_argument( '--res_samples', type = int, default = 64, choices = [ 32, 64 ], help = 'resolution of generated samples' )
    parser.add_argument( '--res_dataset', type = int, default = 64, \
      help = "this is the size of your dataset's shorter dimension (H or W), before any resampling is done;" + \
             " this also places an upper limit on the sample resolution (i.e. the max value that res_samples above can be)" )

    parser.add_argument( '--blur_type', type = str.casefold, default = None, choices = [ None, 'binomial', 'gaussian', 'box' ], \
      help = 'low-pass filter blurring operation to apply after upsampling and before downsampling; these effectively change the upsampling/downsampling operation,' + \
             ' for e.g., setting `binomial` converts nearest neighbor upsampling into bilinear upsampling, setting `box` just average pools' )

    parser.add_argument( '--eps_drift', type = float, default = 0., help = 'drift loss multiplicative coefficient; 0. means no drift loss' )

    parser.add_argument( '--len_latent', type = int, default = 128, help = 'number of elements that compose latent vector z' )

    parser.add_argument( '--nonlinearity', type = str.casefold, default = 'relu', choices = [ 'leaky relu', 'relu', 'tanh' ], \
      help = 'which nonlinearity to use in generator and discriminator' )
    parser.add_argument( '--leakiness', type = float, default = .01, help = 'leakiness of Leaky ReLU if used (otherwise ignored)' )
    # parser.add_argument( '--use_tanh', type = bool, default = True, help = 'whether to use tanh at the end of the generator' )

    parser.add_argument( '--use_equalized_lr', type = str2bool, nargs = '?', const = True, default = False, \
      help = 'whether to use equalized learning rate in generator and discriminator' )

  # ............................................................................ #

  # Progressive GAN/StyleGAN Keyword Arguments:
  elif _model_type in ( 'ProGAN', 'StyleGAN', ):
    parser.add_argument( '--batch_size', type = int, default = BS, help = 'base batch size of training/generated data' )
    parser.add_argument( '--bs_dict', type = dict, default = { 4:BS, 8:BS, 16:BS, 32:BS, 64:BS, 128:BS, 256:BS, 512:BS//2, 1024:BS//4 }, \
      help = 'resolution-dependent batch sizes;' + \
             ' change these based on the memory-requirements of your device; higher resolutions than 1024 are also supported' )

    parser.add_argument( '--num_disc_iters', type = int, default = 1, \
      help = "number of discriminator iterations in one main iteration through generator and discriminator;" + \
             " no need to train discriminator to optimality here (for faster convergence)" )
    parser.add_argument( '--nimg_transition', type = int, default = NIMG_TRANSITION, \
      help = 'number of real images seen by discriminator before transitioning to next fade-in/stabilization phase' )

    parser.add_argument( '--lr_base', type = float, default = .001, help = 'base (max) learning rate; can vary depending on dataset and model' )
    parser.add_argument( '--lr_sched', type = str.casefold, default = 'resolution dependent', \
      choices = [ None, 'resolution dependent', 'linear decay', 'custom' ], help = 'learning rate scheduler type' )

    parser.add_argument( '--beta2', type = float, default = .99, help = 'variance/RMSprop EWMA decay coefficient (default recommended for Adam)' )

    parser.add_argument( '--res_samples', type = int, default = 1024, choices = np.logspace( 2, 16, num = 15, base = 2, dtype = np.uint64 ), \
      help = 'final resoluton of generated samples' )
    parser.add_argument( '--res_dataset', type = int, default = 1024, choices = np.logspace( 2, 16, num = 15, base = 2, dtype = np.uint64 ), \
      help = "this is the size of your dataset's shorter dimension (H or W), before any resampling is done;" + \
             " this also places an upper limit on the sample resolution (i.e. the max value that res_samples above can be)" )

    parser.add_argument( '--blur_type', type = str.casefold, default = 'binomial', choices = [ None, 'binomial', 'gaussian', 'box' ], \
      help = 'low-pass filter blurring operation to apply after upsampling and before downsampling; these effectively change the upsampling/downsampling operation,' + \
             ' for e.g., setting `binomial` converts nearest neighbor upsampling into bilinear upsampling, setting `box` just average pools' )
    parser.add_argument( '--bit_exact_resampling', type = str2bool, nargs = '?', const = True, default = False, \
      help = "whether to use PIL's resampling methods (if `True`) or torch's resampling functions (if `False`) for the real data's skip connection;" + \
             " the torch resampling functions are faster" )

    parser.add_argument( '--eps_drift', type = float, default = .001, help = 'drift loss multiplicative coefficient; 0. means no drift loss' )

    parser.add_argument( '--len_latent', type = int, default = 512, help = 'number of elements that compose latent vector z' )

    parser.add_argument( '--nonlinearity', type = str.casefold, default = 'leaky relu', choices = [ 'leaky relu', 'relu', 'tanh' ], \
      help = 'which nonlinearity to use in generator and discriminator' )
    parser.add_argument( '--leakiness', type = float, default = .2, help = 'leakiness of Leaky ReLU if used (otherwise ignored)' )
    # parser.add_argument( '--use_tanh', type = bool, default = False, help = 'whether to use tanh at the end of the generator' )

    parser.add_argument( '--use_equalized_lr', type = str2bool, nargs = '?', const = True, default = True, \
      help = 'whether to use equalized learning rate in generator and discriminator' )
    parser.add_argument( '--normalize_z', type = str2bool, nargs = '?', const = True, default = True, \
      help = "whether to use pixelwise feature vector normalization on the latent vector z (regardless of --use_pixelnorm's value)" )

    parser.add_argument( '--mbstd_group_size', type = int, default = 4, \
      help = 'group size for minibatch standard deviation layer in the discriminator to help increase variation in generated images;' + \
             ' setting this to -1 instead will not apply this layer at all' )

    parser.add_argument( '--use_ewma_gen', type = str2bool, nargs = '?', const = True, default = True, \
      help = 'whether to compute an additional time-averaged (EWMA) generator that often yields higher-quality results than the original generator' )

    # .......................................................................... #

    # Progressive GAN-specific Keyword Arguments:
    if _model_type is 'ProGAN':
      parser.add_argument( '--num_main_iters', type = int, default = ( NIMG_TRANSITION // BS ) * 20, \
        help = 'number of main iterations of training generator and discriminator' )

      parser.add_argument( '--lr_fctr_dict', type = dict, default = { 4:1., 8:1., 16:1., 32:1., 64:1., 128:1., 256:1., 512:1., 1024:1.5 }, \
        help = 'resolution-dependent multiplicative factor to multiply lr_base by' )

      parser.add_argument( '--init_res', type = int, default = 4, choices = np.logspace( 2, 16, num = 15, base = 2, dtype = np.uint64 ), \
        help = 'starting resolution of Progressive GAN' )

      parser.add_argument( '--use_pixelnorm', type = str2bool, nargs = '?', const = True, default = True, \
        help = 'whether to use pixelwise feature vector normalization in generator' )

    # .......................................................................... #

    # StyleGAN-specific Keyword Arguments:
    elif _model_type is 'StyleGAN':
      parser.add_argument( '--num_main_iters', type = int, default = ( NIMG_TRANSITION // BS ) * 18, \
        help = 'number of main iterations of training generator and discriminator' )

      parser.add_argument( '--lr_fctr_dict', type = dict, default = { 4:1., 8:1., 16:1., 32:1., 64:1., 128:1.5, 256:2., 512:3., 1024:3. }, \
        help = 'resolution-dependent multiplicative factor to multiply lr_base by' )

      parser.add_argument( '--init_res', type = int, default = 8, choices = np.logspace( 2, 16, num = 15, base = 2, dtype = np.uint64 ), \
        help = 'starting resolution of StyleGAN' )

      parser.add_argument( '--len_dlatent', type = int, default = 512, \
        help = 'number of elements that compose disentangled (intermediate) latent vector w' )
      parser.add_argument( '--mapping_num_fcs', type = int, default = 8, \
        help = 'number of fully-connected layers that make up the mapping network' )
      parser.add_argument( '--mapping_lrmul', type = int, default = .01, \
        help = 'multiplicative factor for the learning rate in the mapping network to reduce instability' )

      parser.add_argument( '--use_noise', type = str2bool, nargs = '?', const = True, default = True, \
        help = 'whether to apply noise inputs to generator in order to control stochastic variation in generated images' )
      parser.add_argument( '--use_pixelnorm', type = str2bool, nargs = '?', const = True, default = False, \
        help = 'whether to use pixelwise feature vector normalization in generator' )
      parser.add_argument( '--use_instancenorm', type = str2bool, nargs = '?', const = True, default = True, \
        help = 'whether to use instance normalization in generator (recommended for AdaIN)' )

      # TODO: Implement option to style-mix more than just 2 latent vectors for the mixing regularization
      parser.add_argument( '--pct_mixing_reg', type = float, default = .9, \
        help = 'percent of images generated from 2 latent vectors z instead of 1; 0. means no mixing regularization' )

      parser.add_argument( '--beta_trunc_trick', type = float, default = .995, \
        help = 'decay coefficient for computing EWMA of disentangled latent vector w during training,' + \
               ' which will be used for the truncation trick in disentangled latent space W during evaluation' )
      parser.add_argument( '--psi_trunc_trick', type = float, default = .7, \
        help = 'multiplicative coefficient in range [0,1) for the truncation trick in disentangled latent space W during evaluation;' + \
               ' the smaller the value, the more the truncation' )
      parser.add_argument( '--cutoff_trunc_trick', type = int, default = 4, \
        help = 'final generator block ( = log2(res) - 1 ) to use the computed psi-truncated disentangled latent vector w during evaluation' )

  # ---------------------------------------------------------------------------- #

  config = parser.parse_args( )

  data_config = get_current_configuration( 'data_config', raise_exception = False )

  # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
  # Post-processing:
  # ----------------

  config.dev = torch.device( config.dev )

  if config.pin_memory and config.dev != torch.device( 'cuda' ):
    config = None
    raise ValueError( '--pin_memory should be set to `False` if not using CUDA.' )

  if config.random_seed == -1:
    np.random.seed( None )
    if config.n_gpu > 1:
      torch.cuda.seed_all()  # multi-GPU, ignored if CPU
    else:
      torch.seed()  # single-processor, CUDA or CPU
  elif 0 <= config.random_seed < 2**32:
    np.random.seed( config.random_seed )
    if config.n_gpu > 1:
      torch.cuda.manual_seed_all( config.random_seed )  # multi-GPU, ignored if CPU
    else:
      torch.manual_seed( config.random_seed )  # single-processor, CUDA or CPU
    torch.backends.cudnn.deterministic = True
    if config.enable_cudnn_autotuner:
      config.enable_cudnn_autotuner = False
      print( 'To maintain determinism, the CuDNN auto-tuner has been disabled...' )
  else:
    config = None
    raise ValueError( "--random_seed must either be -1 for random seeding or" + \
                      " be in the range [0,2**32) to accommodate numpy's and" + \
                      " torch's seeding specifications." )

  torch.backends.cudnn.benchmark = config.enable_cudnn_autotuner

  if _model_type in ( 'ProGAN', 'StyleGAN', ):
    _bs = config.batch_size
    config.bs_dict = { 4:_bs, 8:_bs, 16:_bs, 32:_bs, 64:_bs, 128:_bs, 256:_bs, 512:_bs//2, 1024:_bs//4 }
    if config.mbstd_group_size < -1 or not config.mbstd_group_size:
      raise ValueError( "--mbstd_group_size must either be -1 to indicate not applying" + \
                        " minibatch standard deviation or a positive integer indicating" + \
                        " the group size for the minibatch standard deviation layer." )

  config.save_samples_dir.mkdir( parents = True, exist_ok = True )
  config.save_model_dir.mkdir( parents = True, exist_ok = True )

  configs_dir = str( os.path.abspath( os.path.dirname( __file__ ) ) )

  if data_config is not None:
    for n, transform in enumerate( data_config.ds_transforms.transforms ):
      _num_rsz = 0
      # NOTE: messy control flow below
      if ( isinstance( transform, partial ) and transform.func is transforms.Resize ) or \
        isinstance( transform, transforms.Resize ):
        _num_rsz += 1
        if _num_rsz < 2:
          if _model_type == 'ResNet GAN':
            _res = config.res_samples
          elif _model_type in ( 'ProGAN', 'StyleGAN', ):
            _res = config.init_res
          if isinstance( transform, partial ):
            data_config.ds_transforms.transforms[n] = \
              data_config.ds_transforms.transforms[n]( size = ( _res, _res, ) )
          elif data_config.ds_transforms.transforms[n].size[0] != _res:
            data_config.ds_transforms.transforms[n].size = ( _res, _res, )
        else:
          raise RuntimeWarning( 'Warning: More than 1 "Resize" transform found;' + \
                                ' only updated the first "Resize" transform.' )

    with open( configs_dir + '/.data_config.p', 'wb' ) as f:
      pickle.dump( data_config, f, protocol = 3 )

  config.model = _model_type

  # print( config )

  with open( str( Path.home()/'.configs_dir.txt' ), 'wb' ) as f:
    f.write( configs_dir.encode() )

  with open( configs_dir + '/.config.p', 'wb' ) as f:
    pickle.dump( config, f, protocol = 3 )


# -*- coding: UTF-8 -*-

"""Performs initialization for almost all parameters in the neural networks of this package.

Handles multiple kinds of parameter initializations from multiple kinds of
models (ResNet GAN, ProGAN, StyleGAN, etc.).
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

import numpy as np
import torch

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

class Initializer( object ):
  """Initialization class that handles multiple kinds of parameter initializations."""
  def __init__( self, init, init_type = 'default', gain_sq_base = 2., equalized_lr = False ):
    super( Initializer, self ).__init__()

    self.init = init.casefold()
    self.init_type = init_type.casefold()
    self.gain_sq_base = gain_sq_base
    self.equalized_lr = equalized_lr

  @torch.no_grad()
  def get_init_bound_layer( self, tensor, distribution_type, stride = 1 ):
    distribution_type = distribution_type.casefold()
    if distribution_type not in ( 'uniform', 'normal', ):
      raise ValueError( 'Only uniform and normal distributions are supported.' )

    if self.init_type in ( 'default', 'resnet', ):
      fan_in, fan_out = self._calculate_fan_in_fan_out( tensor = tensor, stride = stride )
      std = self._calculate_init_weight_std( fan_in = fan_in, fan_out = fan_out )
    elif self.init_type in ( 'progan', 'stylegan', ):
      fan_in, _ = self._calculate_fan_in_fan_out( tensor = tensor, stride = stride )
      std = self._calculate_init_weight_std( fan_in = fan_in )

    return np.sqrt( 3 )*std if distribution_type == 'uniform' else std

  @torch.no_grad()
  def _calculate_fan_in_fan_out( self, tensor, stride = 1 ):
    dimensions = tensor.dim()
    if dimensions < 2:
      raise ValueError(
        'Fan in and fan out cannot be computed for tensor with fewer than 2 dimensions.'
      )

    fan_out = None
    if dimensions == 2:  # Linear
      fan_in = tensor.size(1)
      if self.init_type not in ( 'progan', 'stylegan', ):
        fan_out = tensor.size(0)
    else:
      receptive_field_size = 1
      if dimensions > 2:
        receptive_field_size = tensor[0][0].numel()
      fan_in = tensor.size(1) * receptive_field_size
      if self.init_type not in ( 'progan', 'stylegan', ):
        fan_out = tensor.size(0) * receptive_field_size / stride**2

    return fan_in, fan_out

  @torch.no_grad()
  def _calculate_init_weight_std( self, fan_in = None, fan_out = None ):
    gain_sq = self.gain_sq_base / 2.
    if fan_out is not None and fan_in is not None:
      fan = fan_in + fan_out
      gain_sq *= 2
    elif fan_in is not None:
      fan = fan_in
    elif fan_out is not None:
      fan = fan_out

    if self.init == 'he': gain_sq = 2. * gain_sq
    elif self.init == 'xavier': gain_sq = 1. * gain_sq

    std = np.sqrt( gain_sq / fan )

    return std


# -*- coding: UTF-8 -*-

"""Base architecture class for non-progressive GANs.
"""

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

from abc import ABC, abstractmethod

import torch
from torch import nn

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

class GAN( nn.Module, ABC ):
  """Base GAN for all non-progressive architectures."""
  def __init__( self, res ):
    super( GAN, self ).__init__()
    self._res = res

  def most_parameters( self, recurse = True, excluded_params:list = [] ):
    """torch.nn.Module.parameters() generator method but with the option to exclude specified parameters."""
    for name, params in self.named_parameters( recurse = recurse ):
      if name not in excluded_params:
        yield params

  @property
  def res( self ):
    return self._res
  
  @res.setter
  def res( self, new_res ):
    message = f'GAN().res cannot be changed, as {self.__class__.__name__} only permits one resolution: {self._res}.'
    raise AttributeError( message )

  @abstractmethod
  def forward( self, x ):
    raise NotImplementedError( 'Can only call `forward` on valid subclasses.' )


# -*- coding: UTF-8 -*-

"""Learner for StyleGANs.

  Typical usage example:

  First configure your desired GAN on the command-line:
    go to root directory...
    $ python config.py stylegan
    $ python data_config.py FFHQ path/to/datasets/ffhq

  Then write a custom script (or use train.py):
    from gan_lab import get_current_configuration
    from gan_lab.utils.data_utils import prepare_dataset, prepare_dataloader
    from gan_lab.stylegan.learner import StyleGANLearner

    # get most recent configurations:
    config = get_current_configuration( 'config' )
    data_config = get_current_configuration( 'data_config' )

    # get DataLoader(s)
    train_ds, valid_ds = prepare_dataset( data_config )
    train_dl, valid_dl, z_valid_dl = prepare_dataloader( config, data_config, train_ds, valid_ds )

    # instantiate StyleGANLearner and train:
    learner = StyleGANLearner( config )
    learner.train( train_dl, valid_dl, z_valid_dl )   # train for config.num_main_iters iterations
    learner.config.num_main_iters = 300000            # this is one example of changing your instantiated learner's configurations
    learner.train( train_dl, valid_dl, z_valid_dl )   # train for another 300000 iterations

Note that the above custom script is just a more flexible alternative to running
train.py (you can, for example, run the above on a Jupyter Notebook). You can
always just run train.py.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from .base import StyleGAN
from .architectures import StyleAddNoise, StyleGenerator
from _int import get_current_configuration, LearnerConfigCopy
from progan.architectures import ProDiscriminator
from progan.learner import ProGANLearner
from utils.latent_utils import gen_rand_latent_vars
from utils.backprop_utils import calc_gp, configure_adam_for_gan
from utils.custom_layers import Conv2dBias, LinearBias

from abc import ABC
import copy
import logging
from pathlib import Path
from functools import partial
from itertools import accumulate
from timeit import default_timer as timer

import numpy as np
from PIL import Image
from indexed import IndexedOrderedDict
import matplotlib.pyplot as plt
plt.rcParams.update( { 'figure.max_open_warning': 0 } )
import torch
from torch import nn
import torch.nn.functional as F
from torchvision import transforms

# from tqdm import tqdm

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

NONREDEFINABLE_ATTRS = ( 'model', 'init_res', 'res_samples', 'res_dataset', 'len_latent',
                         'num_classes', 'class_condition', 'use_auxiliary_classifier',
                         'model_upsample_type', 'model_downsample_type', 'align_corners',
                         'blur_type', 'nonlinearity', 'use_equalized_lr', 'normalize_z',
                         'use_pixelnorm', 'mbstd_group_size', 'use_ewma_gen',
                         'use_instancenorm', 'use_noise', 'pct_mixing_reg',
                         'beta_trunc_trick', 'psi_trunc_trick', 'cutoff_trunc_trick',
                         'len_dlatent', 'mapping_num_fcs', 'mapping_lrmul', )

REDEFINABLE_FROM_LEARNER_ATTRS = ( 'batch_size', 'loss', 'gradient_penalty',
                                   'optimizer', 'lr_sched', 'latent_distribution', )

COMPUTE_EWMA_VIA_HALFLIFE = True
EWMA_SMOOTHING_HALFLIFE = 10.
EWMA_SMOOTHING_BETA = .999

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

class StyleGANLearner( ProGANLearner ):
  """GAN Learner specifically designed for StyleGAN architectures.

  Once instantiated, the StyleGANLearner object's configuration can be changed, but only via
  its self.config attribute (i.e. running 'python config.py [model]' post-instantiation
  will not affect this learner's configuration).
  """
  def __init__( self, config ):
    super( StyleGANLearner, self ).__init__( config )

    if self.model == 'StyleGAN':

      # If you want to change an attribute in an already-instantiated StyleGANLearner's config or data_config,
      # change self.config (below) instead of config and self.data_config (also below) instead of data_config:
      self.config = LearnerConfigCopy( config,
                                       self.__class__.__name__,
                                       NONREDEFINABLE_ATTRS,
                                       REDEFINABLE_FROM_LEARNER_ATTRS )
      # pretrained models loaded later on for evaluation should not require data_config.py to have been run:
      self._is_data_configed = False; self._stats_set = False
      self._update_data_config( raise_exception = False )

      self._latent_distribution = self.config.latent_distribution

      global StyleGAN, StyleGenerator, ProDiscriminator, StyleDiscriminator
      StyleGAN = type( 'StyleGAN', ( nn.Module, ABC, ), dict( StyleGAN.__dict__ ) )
      StyleGAN.reset_state( )

      StyleGenerator = type( 'StyleGenerator', ( StyleGAN, ), dict( StyleGenerator.__dict__ ) )
      self.gen_model = StyleGenerator(
        final_res = self.config.res_samples,
        latent_distribution = self._latent_distribution,
        len_latent = self.config.len_latent,
        len_dlatent = self.config.len_dlatent,
        mapping_num_fcs = self.config.mapping_num_fcs,
        mapping_lrmul = self.config.mapping_lrmul,
        use_instancenorm = self.config.use_instancenorm,
        use_noise = self.config.use_noise,
        upsampler = self.gen_model_upsampler,
        blur_type = self.config.blur_type,
        nl = self.nl,
        num_classes = self.num_classes_gen,
        equalized_lr = self.config.use_equalized_lr,
        normalize_z = self.config.normalize_z,
        use_pixelnorm = self.config.use_pixelnorm,
        pct_mixing_reg = self.config.pct_mixing_reg,
        truncation_trick_params = { 'beta': self.config.beta_trunc_trick,
                                    'psi': self.config.psi_trunc_trick,
                                    'cutoff_stage': self.config.cutoff_trunc_trick }
      )

      # Create `StyleDiscriminator` type object by converting `ProDiscriminator` class into `StyleDiscriminator` class:
      StyleDiscriminator = type( 'StyleDiscriminator', ( StyleGAN, ), dict( ProDiscriminator.__dict__ ) )
      self.disc_model = StyleDiscriminator(
        final_res = self.config.res_samples,
        pooler = self.disc_model_downsampler,
        blur_type = self.config.blur_type,
        nl = self.nl,
        num_classes = self.num_classes_disc,
        equalized_lr = self.config.use_equalized_lr,
        mbstd_group_size = self.config.mbstd_group_size
      )

      # If one wants to start at a higher resolution than 4:
      assert self.config.init_res <= self.config.res_samples
      if self.config.init_res > 4:
        _init_res_log2 = int( np.log2( self.config.init_res ) )
        if float( self.config.init_res ) != 2**_init_res_log2:
          raise ValueError( 'Only resolutions that are powers of 2 are supported.' )
        num_scale_inc = _init_res_log2 - 2
        for _ in range( num_scale_inc ):
          self.gen_model.increase_scale()
          self.disc_model.increase_scale()
        self.gen_model.fade_in_phase = False  # this applies it to both networks simultaneously

      # Generator and Discriminator state data must match:
      assert self.gen_model.cls_base.__dict__ == \
             self.disc_model.cls_base.__dict__

      # Initialize EWMA Generator Model:
      self.gen_model_lagged = None
      if self.config.use_ewma_gen:
        _orig_mode = self.gen_model.training
        self.gen_model.train()
        self.gen_model.to( 'cpu' )
        with torch.no_grad():
          self.gen_model_lagged = copy.deepcopy( self.gen_model )  # for memory efficiency in GPU
          self.gen_model_lagged.to( self.config.metrics_dev )
          self.gen_model_lagged.train( mode = _orig_mode )
        self.gen_model.train( mode = _orig_mode )

      self.gen_model.to( self.config.dev )
      self.disc_model.to( self.config.dev )

      self.batch_size = self.config.bs_dict[ self.gen_model.curr_res ]

      if self.cond_gen:
        self.labels_one_hot_disc = self._tensor( self.batch_size, self.num_classes )
        self.labels_one_hot_gen = self._tensor( self.batch_size * self.config.gen_bs_mult, self.num_classes )

      # Loss Function:
      self._loss = config.loss.casefold()
      self._set_loss( )

      # Gradient Regularizer:
      self.gp_func = partial(
        calc_gp,
        gp_type = self._gradient_penalty,
        nn_disc = self.disc_model,
        lda = self.config.lda,
        gamma = self.config.gamma
      )

      # Optimizer:
      self._set_optimizer( )

      # Epsilon Loss to punish possible outliers from training distribution:
      self.eps = False
      if self.config.eps_drift > 0:
        self.eps = True

      # Print configuration:
      print( '-------- Initialized Model Configuration --------' )
      print( self.config )
      print( '-------------------------------------------------' )
      # print( "  If you would like to alter any of the above configurations,\n" + \
      #        "  please do so via altering your instantiated StyleGANLearner().config's attributes." )
      print( '\n    Ready to train!\n' )

  def _apply_lagged_weights( self, m ):
    # TODO: Include support for other learnable layers such as BatchNorm
    _keys = m.state_dict().keys()
    if isinstance( m, ( nn.Linear, nn.Conv2d, LinearBias, Conv2dBias, ) ):
      if 'weight' in _keys:
        m.weight = nn.Parameter( self.lagged_params.values()[ self._param_tensor_num + 1 ] )
        self._param_tensor_num += 1
      if 'bias' in _keys:
        m.bias = nn.Parameter( self.lagged_params.values()[ self._param_tensor_num + 1 ] )
        self._param_tensor_num += 1
    elif isinstance( m, StyleAddNoise ):
      if 'noise_weight' in _keys:
        m.noise_weight = nn.Parameter( self.lagged_params.values()[ self._param_tensor_num + 1 ] )
        self._param_tensor_num += 1
    elif isinstance( m, StyleGenerator ):
      if 'const_input' in _keys:
        m.const_input = nn.Parameter( self.lagged_params.values()[ 0 ] )
        self._param_tensor_num += 1

  @property
  def latent_distribution( self ):
    return self._latent_distribution

  @latent_distribution.setter
  def latent_distribution( self, new_latent_distribution ):
    new_latent_distribution = new_latent_distribution.casefold()
    self._latent_distribution = new_latent_distribution
    self.gen_model.latent_distribution = new_latent_distribution
    if self.config.use_ewma_gen:
      self.gen_model_lagged.latent_distribution = new_latent_distribution

  # def reset_stylegan_state( self ):
  #   self.gen_model.cls_base.reset_state( )  # this applies to both networks simultaneously

  # .......................................................................... #

  @torch.no_grad()
  def plot_sample( self, z_test, z_mixing = None, style_mixing_stage = None, noise = None, label = None, time_average = True ):
    """Plots and shows 1 sample from input latent code; offers stylemixing and noise input capabilities."""
    if self.ds_mean is None or self.ds_std is None:
      raise ValueError( "This model does not hold any information about your dataset's mean and/or std.\n" + \
                        "Please provide these (either from your current data configuration or from your pretrained model)." )

    for z in ( z_test, z_mixing, ):
      if z is not None:
        if z.dim() == 2:
          if z.shape[0] != 1:
            raise IndexError( 'This method only permits plotting 1 generated sample at a time.' )
        elif z.dim() != 1:
          raise IndexError( 'Incorrect dimensions of input latent vector. Must be either `dim == 1` or `dim == 2`.' )
        if not self.cond_gen:
          if z.shape[-1] != self.config.len_latent:
            message = f"Input latent vector must be of size {self.config.len_latent}."
            raise IndexError( message )
        else:
          if z.shape[-1] != self.config.len_latent + self.num_classes_gen:
            message = f"This is a generator class-conditioned model. So please make sure to append a one-hot encoded vector\n" + \
                      f"of size {self.num_classes_gen} that indicates the to-be generated sample's class to a latent vector of\n" + \
                      f"size {self.config.len_latent}. Total input size must therefore be {self.config.len_latent + self.num_classes_gen}."
            raise IndexError( message )

    # z_test = z_test.to( self.config.dev )
    x_test = \
      self.gen_model_lagged(
        z_test,
        x_mixing = z_mixing,
        style_mixing_stage = style_mixing_stage,
        noise = noise ).squeeze() if time_average else \
      self.gen_model(
        z_test,
        x_mixing = z_mixing,
        style_mixing_stage = style_mixing_stage,
        noise = noise ).squeeze()

    if label is not None:
      print( f'Label Index for Generated Image: {label}' )

    logger = logging.getLogger()
    _old_level = logger.level
    logger.setLevel( 100 )  # ignores potential "clipping input data" warning
    plt.imshow( ( ( ( x_test ) \
                          .cpu().detach() * self.ds_std ) + self.ds_mean ) \
                          .numpy().transpose( 1, 2, 0 ), interpolation = 'none'
    )
    logger.setLevel( _old_level )
    plt.show()

  @torch.no_grad()
  def make_stylemixing_grid( self, zs_sourceb, zs_coarse = [], zs_middle = [], zs_fine = [], labels = None, time_average = True ):
    """Generates style-mixed grid of images, emulating Figure 3 in Karras et al. 2019."""
    assert any( len( zs ) for zs in ( zs_coarse, zs_middle, zs_fine, ) )
    if self.ds_mean is None or self.ds_std is None:
      raise ValueError( "This model does not hold any information about your dataset's mean and/or std.\n" + \
                        "Please provide these (either from your current data configuration or from your pretrained model)." )

    szs = []
    stages = [ 0 ]
    for m, zs in enumerate( ( zs_sourceb, zs_coarse, zs_middle, zs_fine, ) ):
      szs.append( 1 if isinstance( zs, torch.Tensor ) and zs.dim() == 1 else len( zs ) )
      if szs[m]:
        if zs.dim() > 2:
          raise IndexError( 'Incorrect dimensions of input latent vector. Must be either `dim == 1` or `dim == 2`.' )
        else:
          if m == 1:
            stage = 1  # stage for style-mixing of coarse features from source b
            if zs.dim() == 1: zs_coarse.unsqueeze_( dim = 0 )
          elif m == 2:
            stage = 4  # stage for style-mixing of mid-level features from source b
            if zs.dim() == 1: zs_middle.unsqueeze_( dim = 0 )
          elif m == 3:
            stage = 8  # stage for style-mixing of fine-grained features from source b
            if zs.dim() == 1: zs_fine.unsqueeze_( dim = 0 )
          if m: stages.append( stage )
        if not self.cond_gen:
          if zs.shape[1] != self.config.len_latent:
            message = f"Input latent vector must be of size {self.config.len_latent}."
            raise IndexError( message )
        else:
          if zs.shape[1] != self.config.len_latent + self.num_classes_gen:
            message = f"This is a generator class-conditioned model. So please make sure to append a one-hot encoded vector\n" + \
                      f"of size {self.num_classes_gen} that indicates the to-be generated sample's class to a latent vector of\n" + \
                      f"size {self.config.len_latent}. Total input size must therefore be {self.config.len_latent + self.num_classes_gen}."
            raise IndexError( message )
      else:
          stages.append( 0 )

    ncols_tot = 1 + szs[0]; nrows_tot = 1 + sum( szs[1:] )
    cum_szs = [0] + list( accumulate( szs[1:] ) )
    fig = plt.figure( figsize = ( 8. * ( ncols_tot / nrows_tot ), 8. if labels is None else 9., ) )
    axs = fig.subplots( ncols = ncols_tot, nrows = nrows_tot )
    fig.tight_layout( pad = 0 )
    _fctrs = ( ( fig.subplotpars.wspace / 0.47267497603068315 ), ( fig.subplotpars.hspace / 0.28000000000000086 ), )
    gaps = ( ( .002 * _fctrs[0] * ( fig.dpi / fig.get_size_inches().mean() ) ), \
              ( .002 * _fctrs[1] * ( fig.dpi / fig.get_size_inches()[1] ) ), )
    # wspace = fig.subplotpars.wspace
    # hspace = fig.subplotpars.hspace
    wspace = plt.rcParams[ 'figure.subplot.wspace' ] * _fctrs[0]
    hspace = plt.rcParams[ 'figure.subplot.hspace' ] * _fctrs[1]
    start = 0
    logger = logging.getLogger(); _old_level = logger.level; logger.setLevel( 100 )  # ignores potential "clipping input data" warning
    for m, zs in enumerate( ( zs_sourceb, zs_coarse, zs_middle, zs_fine, ) ):
      if szs[m]:
        nrows = 1 if not m else szs[ m ]
        if m: start = row + 1
        for row in range( start, start + nrows ):
          for col in range( ncols_tot ):
            axs[row][col].axis( 'off' )
            axs[row][col].set_aspect( 'equal' )
            l, b, w, h = axs[row][col].get_position().bounds
            l *= ( 1. - wspace ); b *= ( 1. - hspace ); w *= ( 1. + wspace ); h *= ( 1. + hspace )
            if m > 1: b -= ( m - 1 ) *gaps[1]
            if not row and not col:
              axs[row][col].set_position( [ l - gaps[0], b + gaps[1], w, h ] )
            else:
              if ( not row ) != ( not col ):  # xor
                if row:
                  axs[row][col].set_position( [ l - gaps[0], b, w, h ] )
                  if row == 1:
                    axs[row][col].set_title( 'Source A', fontweight = 'bold' )
                elif col:
                  axs[row][col].set_position( [ l, b + gaps[1], w, h ] )
                  if col == 1:
                    axs[row][col].axis( 'on' )
                    axs[row][col].spines['top'].set_visible(False)
                    axs[row][col].spines['right'].set_visible(False)
                    axs[row][col].spines['bottom'].set_visible(False)
                    axs[row][col].spines['left'].set_visible(False)
                    axs[row][col].set_xticks( [] )
                    axs[row][col].set_yticks( [] )
                    axs[row][col].set_ylabel( 'Source B', fontweight = 'bold', fontsize = plt.rcParams[ 'axes.titlesize' ] )
                x = self.gen_model_lagged( zs_sourceb[ col - 1 ] if not m else zs[ row - cum_szs[m-1] - 1 ] ).squeeze() if time_average else \
                    self.gen_model( zs_sourceb[ col - 1 ] if not m else zs[ row - cum_szs[m-1] - 1 ] ).squeeze()
              elif row and col:
                axs[row][col].set_position( [ l, b, w, h ] )
                x = self.gen_model_lagged( zs[ row - cum_szs[m-1] - 1 ], x_mixing = zs_sourceb[ col - 1 ], style_mixing_stage = stages[ m ] ).squeeze() if time_average else \
                    self.gen_model( zs[ row - cum_szs[m-1] - 1 ], x_mixing = zs_sourceb[ col - 1 ], style_mixing_stage = stages[ m ] ).squeeze()
              axs[row][col].imshow(
                ( ( x.cpu().detach() * self.ds_std ) + self.ds_mean ).numpy().transpose( 1, 2, 0 ), interpolation = 'none'
              )
              if labels is not None:
                axs[row][col].set_title( str( labels[ row*ncols_tot + col ].item() ) )

    # center the axes of each subplot on the figure after moving everything around
    fig_top_right = fig.get_window_extent().corners()[3]
    l,b,w,h = axs[0,-1].get_position().bounds; top_right = ( l + w, b + h, )
    l,b,w,h = axs[-1,0].get_position().bounds; bot_left = ( l, b, )
    shift = tuple( -bot_left[i] + ( 1. - top_right[i] ) / 2. for i in range( 2 ) )
    for row in range( nrows_tot ):
      for col in range( ncols_tot ):
        l, b, w, h = axs[row][col].get_position().bounds
        axs[row][col].set_position( [ l + shift[0], b + shift[1], w, h ] )

    # ylabel for Coarse, Middle, and Fine styles from Source B
    for k in range( 1, m + 1 ):
      if szs[k]:
        l, b, w, h = axs[cum_szs[k]][0].get_position().bounds
        center_y = ( szs[k] * h ) / 2.
        if k == 1: ylabel = 'Coarse'
        elif k == 2: ylabel = 'Middle'
        elif k == 3: ylabel = 'Fine'
        ylabel += ' from B' if szs[k] < 2 else ' styles from Source B'
        fig.text( l - gaps[0], b + center_y, ylabel, va = 'center', rotation = 'vertical' )
    logger.setLevel( _old_level )

    return ( fig, axs, )

  # .......................................................................... #

  def save_model( self, save_path:Path ):
    if self.not_trained_yet:
      raise Exception( 'Please train your model for atleast 1 iteration before saving.' )

    self.gen_model_metadata = { 'gen_model_upsampler': self.gen_model_upsampler,
                                'num_classes_gen': self.num_classes_gen }
    self.disc_model_metadata = { 'disc_model_downsampler': self.disc_model_downsampler,
                                 'num_classes_disc': self.num_classes_disc }

    scheduler_state_dict_save_args = {
      'scheduler_gen_state_dict': self.scheduler_gen.state_dict() if self.sched_bool else None,
      'scheduler_disc_state_dict': self.scheduler_disc.state_dict() if self.sched_bool else None
    }

    save_path = Path( save_path )
    save_path.parents[0].mkdir( parents = True, exist_ok = True )

    torch.save( {
                'config': self.config,
                'curr_res': self.gen_model.curr_res,
                'alpha': self.gen_model.alpha,
                'use_truncation_trick': self.gen_model.use_truncation_trick,
                'trunc_cutoff_stage': self.gen_model.trunc_cutoff_stage,
                'w_eval_psi': self.gen_model.w_eval_psi,
                'w_ewma_beta': self.gen_model.w_ewma_beta,
                'w_ewma': self.gen_model.w_ewma.to( 'cpu' ),
                'w_ewma_lagged': self.gen_model_lagged.w_ewma.to( 'cpu' ) if self.config.use_ewma_gen else None,
                'trained_with_noise': self.gen_model._trained_with_noise,
                'pct_mixing_reg': self.gen_model.pct_mixing_reg,
                'gen_model_metadata': self.gen_model_metadata,
                'gen_model_state_dict': self.gen_model.state_dict(),
                'gen_model_lagged_state_dict': self.gen_model_lagged.state_dict() if self.config.use_ewma_gen else None,
                'disc_model_metadata': self.disc_model_metadata,
                'disc_model_state_dict': self.disc_model.state_dict(),
                'nl': self.nl,
                'sched_stop_step': self.sched_stop_step,
                'lr_sched': self.lr_sched,
                **scheduler_state_dict_save_args,
                'optimizer': self.optimizer,
                'opt_gen_state_dict': self.opt_gen.state_dict(),
                'opt_disc_state_dict': self.opt_disc.state_dict(),
                'loss': self.loss,
                'gradient_penalty': self.gradient_penalty,
                'batch_size': self.batch_size,
                'curr_dataset_batch_num': self.curr_dataset_batch_num,
                'curr_epoch_num': self.curr_epoch_num,
                'curr_valid_num': self.curr_valid_num,
                'dataset_sz': self.dataset_sz,
                'ac': self.ac,
                'cond_gen': self.cond_gen,
                'cond_disc': self.cond_disc,
                'valid_z': self.valid_z.to( 'cpu' ),
                'valid_label': self.valid_label,
                'grid_inputs_constructed': self.grid_inputs_constructed,
                'rand_idxs': self.rand_idxs,
                'gen_metrics_num': self.gen_metrics_num,
                'disc_metrics_num': self.disc_metrics_num,
                'curr_img_num': self.curr_img_num,
                'nimg_transition_lst': self.nimg_transition_lst,
                'not_trained_yet': self.not_trained_yet,
                'ds_mean': self.ds_mean,
                'ds_std': self.ds_std,
                'latent_distribution': self.latent_distribution,
                'curr_phase_num': self.curr_phase_num,
                'lagged_params': self.lagged_params,
                'progressively_grow': self.progressively_grow }, save_path
    )

  def load_model( self, load_path, dev_of_saved_model = 'cpu' ):
    dev_of_saved_model = dev_of_saved_model.casefold()
    assert ( dev_of_saved_model == 'cpu' or dev_of_saved_model == 'cuda' )
    _map_location = lambda storage,loc: storage if dev_of_saved_model == 'cpu' else None

    checkpoint = torch.load( load_path, map_location = _map_location )

    self.config = checkpoint[ 'config' ]

    self.nl = checkpoint[ 'nl' ]
    self._latent_distribution = checkpoint[ 'latent_distribution' ]

    # Load pretrained neural networks:
    global StyleGAN, StyleGenerator, ProDiscriminator, StyleDiscriminator
    StyleGAN = type( 'StyleGAN', ( nn.Module, ABC, ), dict( StyleGAN.__dict__ ) )
    StyleGAN.reset_state( )

    StyleGenerator = type( 'StyleGenerator', ( StyleGAN, ), dict( StyleGenerator.__dict__ ) )
    self.gen_model_metadata = checkpoint[ 'gen_model_metadata' ]
    self.gen_model_upsampler = self.gen_model_metadata[ 'gen_model_upsampler' ]
    self.num_classes_gen = self.gen_model_metadata[ 'num_classes_gen' ]
    self.gen_model = StyleGenerator(
      final_res = self.config.res_samples,
      latent_distribution = self._latent_distribution,
      len_latent = self.config.len_latent,
      len_dlatent = self.config.len_dlatent,
      mapping_num_fcs = self.config.mapping_num_fcs,
      mapping_lrmul = self.config.mapping_lrmul,
      use_instancenorm = self.config.use_instancenorm,
      use_noise = self.config.use_noise,
      upsampler = self.gen_model_upsampler,
      blur_type = self.config.blur_type,
      nl = self.nl,
      num_classes = self.num_classes_gen,
      equalized_lr = self.config.use_equalized_lr,
      normalize_z = self.config.normalize_z,
      use_pixelnorm = self.config.use_pixelnorm,
      pct_mixing_reg = self.config.pct_mixing_reg,
      truncation_trick_params = { 'beta': self.config.beta_trunc_trick,
                                  'psi': self.config.psi_trunc_trick,
                                  'cutoff_stage': self.config.cutoff_trunc_trick }
    )
    self.gen_model.w_ewma_beta = checkpoint[ 'w_ewma_beta' ]
    self.gen_model._w_eval_psi = checkpoint[ 'w_eval_psi' ]
    self.gen_model._trunc_cutoff_stage = checkpoint[ 'trunc_cutoff_stage' ]
    self.gen_model.use_truncation_trick = checkpoint[ 'use_truncation_trick' ]
    self.gen_model.w_ewma = checkpoint[ 'w_ewma' ]
    self.gen_model._trained_with_noise = checkpoint[ 'trained_with_noise' ]
    self.gen_model._use_noise = checkpoint[ 'trained_with_noise' ]
    self.gen_model.pct_mixing_reg = checkpoint[ 'pct_mixing_reg' ]
    self.gen_model._use_mixing_reg = True if self.gen_model.pct_mixing_reg else False

    # Create `StyleDiscriminator` type object by converting `ProDiscriminator` class into `StyleDiscriminator` class:
    StyleDiscriminator = type( 'StyleDiscriminator', ( StyleGAN, ), dict( ProDiscriminator.__dict__ ) )
    self.disc_model_metadata = checkpoint[ 'disc_model_metadata' ]
    self.disc_model_downsampler = self.disc_model_metadata[ 'disc_model_downsampler' ]
    self.num_classes_disc = self.disc_model_metadata[ 'num_classes_disc' ]
    self.disc_model = StyleDiscriminator(
      final_res = self.config.res_samples,
      pooler = self.disc_model_downsampler,
      blur_type = self.config.blur_type,
      nl = self.nl,
      num_classes = self.num_classes_disc,
      equalized_lr = self.config.use_equalized_lr,
      mbstd_group_size = self.config.mbstd_group_size
    )

    assert self.config.init_res <= self.config.res_samples

    # If pretrained model started at a higher resolution than 4:
    _curr_res = checkpoint[ 'curr_res' ]
    if _curr_res > 4:
      _init_res_log2 = int( np.log2( _curr_res ) )
      if float( _curr_res ) != 2**_init_res_log2:
        raise ValueError( 'Only resolutions that are powers of 2 are supported.' )
      num_scale_inc = _init_res_log2 - 2
      for _ in range( num_scale_inc ):
        self.gen_model.increase_scale()
        self.disc_model.increase_scale()

    # this applies to both networks simultaneously and takes care of fade_in_phase
    self.gen_model.alpha = checkpoint[ 'alpha' ]

    # Generator and Discriminator state data must match:
    assert self.gen_model.cls_base.__dict__ == \
           self.disc_model.cls_base.__dict__

    # Initialize EWMA Generator Model:
    self.gen_model_lagged = None
    if self.config.use_ewma_gen:
      _orig_mode = self.gen_model.training
      self.gen_model.train()
      self.gen_model.to( 'cpu' )
      with torch.no_grad():
        self.gen_model_lagged = copy.deepcopy( self.gen_model )  # for memory efficiency in GPU
        self.gen_model_lagged.to( self.config.metrics_dev )
        self.gen_model_lagged.train( mode = _orig_mode )
      self.gen_model.train( mode = _orig_mode )

    self.gen_model.to( self.config.dev )
    self.gen_model.load_state_dict( checkpoint[ 'gen_model_state_dict' ] )
    self.gen_model.zero_grad()
    if self.config.use_ewma_gen:
      self.gen_model_lagged.load_state_dict( checkpoint[ 'gen_model_lagged_state_dict' ] )
      self.gen_model.w_ewma = checkpoint[ 'w_ewma_lagged' ]
      self.gen_model_lagged.to( self.config.metrics_dev )
      self.gen_model_lagged.zero_grad()
    self.disc_model.to( self.config.dev )
    self.disc_model.load_state_dict( checkpoint[ 'disc_model_state_dict' ] )
    self.disc_model.zero_grad()

    self.sched_stop_step = checkpoint[ 'sched_stop_step' ]
    self.lr_sched = checkpoint[ 'lr_sched' ]
    self._scheduler_gen_state_dict = checkpoint['scheduler_gen_state_dict']
    self._scheduler_disc_state_dict = checkpoint['scheduler_disc_state_dict']
    self._sched_state_dict_set = False

    self.optimizer = checkpoint['optimizer']
    self.opt_gen.load_state_dict( checkpoint['opt_gen_state_dict'] )
    self.opt_disc.load_state_dict( checkpoint['opt_disc_state_dict'] )

    self.batch_size = checkpoint['batch_size']

    self.loss = checkpoint['loss']

    self.gradient_penalty = checkpoint['gradient_penalty']

    self.curr_dataset_batch_num = checkpoint['curr_dataset_batch_num']
    self.curr_epoch_num = checkpoint['curr_epoch_num']
    self.curr_valid_num = checkpoint['curr_valid_num']
    self.dataset_sz = checkpoint['dataset_sz']

    self.ac = checkpoint['ac']
    self.cond_gen = checkpoint['cond_gen']
    self._tensor = torch.FloatTensor
    if self.config.dev == torch.device( 'cuda' ):
      self._tensor = torch.cuda.FloatTensor
    if self.cond_gen:
      self.labels_one_hot_disc = self._tensor( self.batch_size, self.num_classes )
      self.labels_one_hot_gen = self._tensor( self.batch_size * self.config.gen_bs_mult, self.num_classes )
    self.cond_disc = checkpoint['cond_disc']

    self.valid_z = checkpoint['valid_z'].to( self.config.metrics_dev )
    self.valid_label = checkpoint['valid_label']
    if self.valid_label is not None:
      self.valid_label = self.valid_label.to( 'cpu' )
    self.grid_inputs_constructed = checkpoint['grid_inputs_constructed']
    self.rand_idxs = checkpoint['rand_idxs']
    self.gen_metrics_num = checkpoint['gen_metrics_num']
    self.disc_metrics_num = checkpoint['disc_metrics_num']

    self.curr_img_num = checkpoint[ 'curr_img_num' ]
    self.nimg_transition_lst = checkpoint[ 'nimg_transition_lst' ]

    self.not_trained_yet = checkpoint['not_trained_yet']

    # By default, use the pretrained model's statistics (you can change this by changing ds_mean and ds_std manually)
    self.ds_mean = checkpoint['ds_mean']
    self.ds_std = checkpoint['ds_std']
    if self._is_data_configed:
      self.data_config.ds_mean = self.ds_mean.squeeze().tolist()
      self.data_config.ds_std = self.ds_std.squeeze().tolist()

    self.latent_distribution = checkpoint[ 'latent_distribution' ]
    self.curr_phase_num = checkpoint[ 'curr_phase_num' ]
    self.eps = False
    if self.config.eps_drift > 0:
      self.eps = True
    self.lagged_params = checkpoint[ 'lagged_params' ]
    self._progressively_grow = checkpoint[ 'progressively_grow' ]


    self.pretrained_model = True

    # Print configuration:
    print( '---------- Loaded Model Configuration -----------' )
    print( self.config )
    print( '-------------------------------------------------' )
    print( "\n  If you would like to change any of the above configurations,\n" + \
           "  please do so via setting the attributes of your instantiated StyleGANLearner().config object.\n" )


# -*- coding: UTF-8 -*-

"""ResNet architectures (generators and discriminators).
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from .base import GAN
from .resblocks import ResBlock2d, ResBlock2d32Pix, FastResBlock2dDownsample
from _int import FMAP_SAMPLES, RES_INIT
from utils.custom_layers import Lambda, NormalizeLayer, Conv2dEx, LinearEx

import torch
from torch import nn

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

FMAP_G = 64
FMAP_D = 64
FMAP_G_INIT_32_FCTR = 1
FMAP_G_INIT_64_FCTR = 4

RES_FEATURE_SPACE = 4

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Generators:
# -----------

class Generator32PixResnet( GAN ):
  """ResNet GAN Generator for 32-pixel samples with optional class-conditioning (Mirza & Osindero, 2014).
  """
  def __init__( self, len_latent = 128, fmap = FMAP_G*2, upsampler = nn.Upsample( scale_factor = 2, mode = 'nearest' ),
                blur_type = None, nl = nn.ReLU(), num_classes = 0, equalized_lr = False ):
    super( Generator32PixResnet, self ).__init__( 32 )

    self.len_latent = len_latent
    self.num_classes = num_classes

    self.equalized_lr = equalized_lr

    _fmap_init_32 = len_latent * FMAP_G_INIT_32_FCTR
    self.generator_model = nn.Sequential(
      Lambda( lambda x: x.view( -1, len_latent + num_classes ) ),
      LinearEx( nin_feat = len_latent + num_classes, nout_feat = _fmap_init_32 * RES_INIT**2,
                init = 'Xavier', equalized_lr = equalized_lr ),
      Lambda( lambda x: x.view( -1, _fmap_init_32, RES_INIT, RES_INIT ) ),
      ResBlock2d32Pix( ni = _fmap_init_32, nf = fmap, ks = 3, norm_type = 'BatchNorm', upsampler = upsampler, \
        init = 'He', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d32Pix( ni = fmap, nf = fmap, ks = 3, norm_type = 'BatchNorm', upsampler = upsampler, \
        init = 'He', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d32Pix( ni = fmap, nf = fmap, ks = 3, norm_type = 'BatchNorm', upsampler = upsampler, \
        init = 'He', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type ),
      NormalizeLayer( 'BatchNorm', ni = fmap ),
      nl,
      Conv2dEx( ni = fmap, nf = FMAP_SAMPLES, ks = 3, stride = 1, padding = 1, init = 'Xavier', equalized_lr = equalized_lr ),
      nn.Tanh()
    )

  def forward( self, x ):
    return self.generator_model( x )


class Generator64PixResnet( GAN ):
  """ResNet GAN Generator for 64-pixel samples with optional class-conditioning (Mirza & Osindero, 2014).
  """
  def __init__( self, len_latent = 128, fmap = FMAP_G, upsampler = nn.Upsample( scale_factor = 2, mode = 'nearest' ),
                blur_type = None, nl = nn.ReLU(), num_classes = 0, equalized_lr = False ):
    super( Generator64PixResnet, self ).__init__( 64 )

    self.len_latent = len_latent
    self.num_classes = num_classes

    self.equalized_lr = equalized_lr

    _fmap_init_64 = len_latent * FMAP_G_INIT_64_FCTR
    self.generator_model = nn.Sequential(
      Lambda( lambda x: x.view( -1, len_latent + num_classes ) ),
      LinearEx( nin_feat = len_latent + num_classes, nout_feat = _fmap_init_64 * RES_INIT**2,
                init = 'Xavier', equalized_lr = equalized_lr ),
      Lambda( lambda x: x.view( -1, _fmap_init_64, RES_INIT, RES_INIT ) ),
      ResBlock2d( ni = _fmap_init_64, nf = 8*fmap, ks = 3, norm_type = 'BatchNorm', upsampler = upsampler, \
        init = 'He', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d( ni = 8*fmap, nf = 4*fmap, ks = 3, norm_type = 'BatchNorm', upsampler = upsampler, \
        init = 'He', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d( ni = 4*fmap, nf = 2*fmap, ks = 3, norm_type = 'BatchNorm', upsampler = upsampler, \
        init = 'He', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d( ni = 2*fmap, nf = 1*fmap, ks = 3, norm_type = 'BatchNorm', upsampler = upsampler, \
        init = 'He', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type ),
      NormalizeLayer( 'BatchNorm', ni = 1*fmap ),
      nl,
      Conv2dEx( ni = 1*fmap, nf = FMAP_SAMPLES, ks = 3, stride = 1, padding = 1, init = 'He', equalized_lr = equalized_lr ),
      nn.Tanh()
    )

  def forward( self, x ):
    return self.generator_model( x )

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Discriminators:
# ---------------

class Discriminator32PixResnet( GAN ):
  """ResNet GAN Discriminator for 32-pixel samples with optional class-conditioning (Mirza & Osindero, 2014).
  """
  def __init__( self, fmap = FMAP_D*2, pooler = nn.AvgPool2d( kernel_size = 2, stride = 2 ),
                blur_type = None, nl = nn.ReLU(), num_classes = 0, equalized_lr = False ):
    super( Discriminator32PixResnet, self ).__init__( 32 )

    self.num_classes = num_classes
    self.equalized_lr = equalized_lr

    self.view1 = Lambda( lambda x: x.view( -1, FMAP_SAMPLES + num_classes, self.res, self.res ) )
    self.conv1 = FastResBlock2dDownsample( ni = FMAP_SAMPLES + num_classes, nf = fmap, ks = 3, pooler = pooler,
                                           init = 'Xavier', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type )
    self.resblocks = nn.Sequential(
      ResBlock2d32Pix( ni = fmap, nf = fmap, ks = 3, norm_type = 'LayerNorm', pooler = pooler, \
        init = 'He', nl = nl, res = self.res//2, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d32Pix( ni = fmap, nf = fmap, ks = 3, norm_type = 'LayerNorm', \
        init = 'He', nl = nl, res = self.res//4, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d32Pix( ni = fmap, nf = fmap, ks = 3, norm_type = 'LayerNorm', \
        init = 'He', nl = nl, res = self.res//4, equalized_lr = equalized_lr, blur_type = blur_type ),
      nl,
      nn.AvgPool2d( kernel_size = self.res//4, stride = self.res//4 ),
      Lambda( lambda x: x.view( -1, fmap ) )  # final feature space
      # NormalizeLayer( 'LayerNorm', ni = fmap, res = 1 )
    )
    self.linear1 = LinearEx( nin_feat = fmap, nout_feat = 1, init = 'Xavier', equalized_lr = equalized_lr )

  def forward( self, x ):
    return ( self.linear1( self.resblocks( self.conv1( self.view1( x ) ) ) ) ).view( -1 )


class DiscriminatorAC32PixResnet( Discriminator32PixResnet ):
  """ResNet GAN Discriminator for 32-pixel samples with auxiliary classifier and discriminator class-conditioning removed.
  """
  def __init__( self, fmap = FMAP_D*2, pooler = nn.AvgPool2d( kernel_size = 2, stride = 2 ),
                blur_type = None, nl = nn.ReLU(), num_classes = 0, equalized_lr = False ):
    super( DiscriminatorAC32PixResnet, self ).__init__( fmap, pooler, nl, num_classes, equalized_lr, blur_type )

    self.view1 = Lambda( lambda x: x.view( -1, FMAP_SAMPLES, self.res, self.res ) )
    self.conv1 = FastResBlock2dDownsample( ni = FMAP_SAMPLES, nf = fmap, ks = 3, pooler = pooler,
                                           init = 'Xavier', nl = nl, equalized_lr = equalized_lr, blur_type = blur_type )
    self.linear_aux = LinearEx( nin_feat = fmap, nout_feat = num_classes, init = 'Xavier', equalized_lr = equalized_lr )

  def forward( self, x ):
    x = self.resblocks( self.conv1( self.view1( x ) ) )
    return ( self.linear1( x ) ).view( -1 ), self.linear_aux( x )


class Discriminator64PixResnet( GAN ):
  """ResNet GAN Discriminator for 64-pixel samples with optional class-conditioning (Mirza & Osindero, 2014).
  """
  def __init__( self, fmap = FMAP_D, pooler = nn.AvgPool2d( kernel_size = 2, stride = 2 ),
                blur_type = None, nl = nn.ReLU(), num_classes = 0, equalized_lr = False ):
    super( Discriminator64PixResnet, self ).__init__( 64 )

    self.num_classes = num_classes
    self.equalized_lr = equalized_lr

    self.view1 = Lambda( lambda x: x.view( -1, FMAP_SAMPLES + num_classes, self.res, self.res ) )
    self.conv1 = Conv2dEx( ni = FMAP_SAMPLES + num_classes, nf = 1*fmap, ks = 3, stride = 1,
                           padding = 1, init = 'Xavier', equalized_lr = equalized_lr )
    self.resblocks = nn.Sequential(
      ResBlock2d( ni = 1*fmap, nf = 2*fmap, ks = 3, norm_type = 'LayerNorm', pooler = pooler, \
        init = 'He', nl = nl, res = self.res//1, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d( ni = 2*fmap, nf = 4*fmap, ks = 3, norm_type = 'LayerNorm', pooler = pooler, \
        init = 'He', nl = nl, res = self.res//2, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d( ni = 4*fmap, nf = 8*fmap, ks = 3, norm_type = 'LayerNorm', pooler = pooler, \
        init = 'He', nl = nl, res = self.res//4, equalized_lr = equalized_lr, blur_type = blur_type ),
      ResBlock2d( ni = 8*fmap, nf = 8*fmap, ks = 3, norm_type = 'LayerNorm', pooler = pooler, \
        init = 'He', nl = nl, res = self.res//8, equalized_lr = equalized_lr, blur_type = blur_type ),
      Lambda( lambda x: x.view( -1, RES_FEATURE_SPACE**2 * 8*fmap ) )  # final feature space
      # NormalizeLayer( 'LayerNorm', ni = fmap, res = 1 )
    )
    self.linear1 = LinearEx( nin_feat = RES_FEATURE_SPACE**2 * 8*fmap, nout_feat = 1,
                             init = 'Xavier', equalized_lr = equalized_lr )

  def forward( self, x ):
    return ( self.linear1( self.resblocks( self.conv1( self.view1( x ) ) ) ) ).view( -1 )


class DiscriminatorAC64PixResnet( Discriminator64PixResnet ):
  """ResNet GAN Discriminator for 64-pixel samples with auxiliary classifier and discriminator class-conditioning removed.
  """
  def __init__( self, fmap = FMAP_D, pooler = nn.AvgPool2d( kernel_size = 2, stride = 2 ),
                blur_type = None, nl = nn.ReLU(), num_classes = 0, equalized_lr = False ):
    super( DiscriminatorAC64PixResnet, self ).__init__( fmap, pooler, nl, num_classes, equalized_lr, blur_type )

    self.view1 = Lambda( lambda x: x.view( -1, FMAP_SAMPLES, self.res, self.res ) )
    self.conv1 = Conv2dEx( ni = FMAP_SAMPLES, nf = 1*fmap, ks = 3, stride = 1,
                           padding = 1, init = 'Xavier', equalized_lr = equalized_lr )
    self.linear_aux = LinearEx( nin_feat = RES_FEATURE_SPACE**2 * 8*fmap, nout_feat = num_classes,
                                init = 'Xavier', equalized_lr = equalized_lr )

  def forward( self, x ):
    x = self.resblocks( self.conv1( self.view1( x ) ) )
    return ( self.linear1( x ) ).view( -1 ), self.linear_aux( x )


# -*- coding: UTF-8 -*-

"""Progressive Growing of GANs for Improved Quality, Stability, and Variation.
   (Karras et al. 2018)

Default configuration exactly emulates the most recent official implementation.

That being said, there are plenty of hyperparameter options/settings that you
can change in config.py to alter your ProGAN to your liking.
"""

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

import os; import sys
sys.path.append(
  os.path.abspath( os.path.join( os.path.dirname( __file__ ), '..' ) )
)

from . import base
from . import architectures
from . import learner

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#